 ---------------------------------------------------------------------------
--                       Paul Scherrer Institute (PSI)
-- ---------------------------------------------------------------------------
-- Unit    : latency_measurement.vhd
-- Author  : Jonas Purtschert
-- ---------------------------------------------------------------------------
-- Copyright© PSI, Section DSV
-- ---------------------------------------------------------------------------
-- Comment : Latency Measurement for IFC1210 Interrupt latency debugging
-- ---------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.math_real.all;

use work.evr320_pkg.all;

entity latency_measurement is
  generic (
    CLK_FREQ_HZ : natural := 125000000 -- Xuser Clk Frequency in Hz
  );
  port (
    evr_clk_i             : in std_logic;
    xuser_clk_i           : in std_logic;
    decoder_event_valid_i : in std_logic;
    decoder_event_i       : in std_logic_vector(7 downto 0);
    status_o              : out typ_rec_latency_measure_stat;
    ctrl_i                : in typ_rec_latency_measure_ctrl
  );
end latency_measurement;

architecture rtl of latency_measurement is

  -- --------------------------------------------------------------------------
  -- Signal, Types, Constants
  -- --------------------------------------------------------------------------
    type state_type is (armed, count);
    signal state                   : state_type;
    signal counter                 : unsigned(31 downto 0);
    signal event_nr_sync, event_nr : std_logic_vector(7 downto 0);
    signal event_detected          : std_logic_vector(3 downto 0);
    signal event_detected_sync     : std_logic_vector(1 downto 0);
    constant MAX_COUNT             : unsigned(31 downto 0) := to_unsigned(CLK_FREQ_HZ / 100, 32); -- MAX 10ms

begin

    -- Process: filter events for matching event_nr register:
    ---------------------------------------------------------
    ext_event_proc : process(evr_clk_i)
    begin
      if (rising_edge(evr_clk_i)) then
        -- sync to MGT clock domain:
        event_nr_sync <= ctrl_i.event_nr;
        event_nr      <= event_nr_sync;

        -- check if event has been detected and stretch pulse: 
        event_detected <= event_detected(2 downto 0) & '0';
        if (decoder_event_valid_i = '1' and decoder_event_i = event_nr) then
          event_detected <= (others => '1');
        end if;
      end if;
    end process;

    -- Process: Counter when configured event has been detected:
    ------------------------------------------------------------
    lat_meas_proc : process(xuser_clk_i, counter)
    begin
      if rising_edge(xuser_clk_i) then
        -- sync to user clock domain:
        event_detected_sync <= event_detected_sync(0) & event_detected(3);

        -- counter FSM:
        ---------------
        case state is
          -- counter is armed:
          when armed =>
            counter <= (others => '0');
            -- start counting when event detected (rising edge):
            if (event_detected_sync(1) = '0' and event_detected_sync(0) = '1') then
              state <= count;
            end if;

          -- counting:
          when count =>
            -- count only up to 10ms, and stop:
            if (counter < MAX_COUNT) then
              counter <= counter + 1;
            end if;
            if (ctrl_i.counter_arm = '1') then
              state <= armed;
            end if;
        end case;
      end if;
      status_o.counter_val <= std_logic_vector(counter);
      status_o.event_detected <= event_detected_sync(event_detected_sync'left);
    end process;

end rtl;
